#include "fitz.h"
#include "dev_epub.h"

#define LINEDIST 0.9f
#define SPACEDIST 0.2f

#undef DEBUG_MATCH

#ifdef DEBUG_MATCH
#define DBUG(A) (printf A)
#else
#define DBUG(A) do {} while (0==1)
#endif


#include <ft2build.h>
#include FT_FREETYPE_H

#if ((FREETYPE_MAJOR == 2) && (FREETYPE_MINOR == 1)) || \
	((FREETYPE_MAJOR == 2) && (FREETYPE_MINOR == 2)) || \
	((FREETYPE_MAJOR == 2) && (FREETYPE_MINOR == 3) && (FREETYPE_PATCH < 8))

int
FT_Get_Advance(FT_Face face, int gid, int masks, FT_Fixed *out)
{
	int err;
	err = FT_Load_Glyph(face, gid, masks);
	if (err)
		return err;
	if (masks & FT_LOAD_VERTICAL_LAYOUT)
		*out = face->glyph->metrics.vertAdvance << 10;
	else
		*out = face->glyph->metrics.horiAdvance << 10;
	return 0;
}

#else
#include FT_ADVANCES_H
#endif

/*
 * Using bbox is no good to us, as we need to cope with rotated/flipped
 * glyphs etc. Instead we use an rbox:
 *
 *         <--w-->        X = origin (on baseline of glyph)
 *     ^    ******        h = height of glyph
 *     |   *     *        d = descent of glyph
 *    h|   *     *        w = width (really 'advance')
 *     |    ******        dir = normalised direction of writing (vector)
 *     +   X     *
 *    d|    *    *        (and, not stored, but used for exposition purposes
 *     v     ****         pdir = dir rotated 90 degrees clockwise)
 *         --dir->
 *
 *
 *  ^  *******    ^          ^     *X*     |
 *  |  *   *  *   |          |  *********  |
 *  w  *   *  *  dir         |      *      |
 *  |  *   *  *   |          w    *****   dir
 *  v   *** X*    |          |      *      |
 *     <-h--+d>              v    ** **    v
 *                              <-d-+-h->
 *
 * h is always measured in -pdir and d in pdir.
 *
 * For horizontal writing styles, dir is (1, 0).
 * PDF only really copes with left to right motion: w, h and d are all
 * positive (normally).
 *
 * For languages that move right to left (such as hebrew), either the
 * glyphs will be emitted left -> right (in which case we will end up
 * collating them the wrong way), or they will emit them right to left with
 * large corrections between each glyph (and we'll still end up collating
 * them the wrong way).
 *
 * The 'expected origin' of the successive glyph is therefore
 * origin + w * dir.
 *
 * Given this definition, the 'expected baseline' of the next line should
 * pass through origin + SIZE * pdir
 *
 * For vertical writing styles, dir is (0, -1).
 * PDF only really copes with top to bottom motion: w is positive (normally).
 * h and d are all positive (normally).
 *
 * For languages that output bottom to top (are there any?), either the
 * glyphs will be emitted top to bottom, or they will emit them bottom to
 * top with large corrections between each glyph. In either case we'll end
 * up collating them in reverse.
 *
 * The 'expected origin' of successive glyphs is given as above.
 * Most vertical writing systems move right to left for successive columns.
 * This fits with our 'expected baseline' equation above too.
 */

struct fz_epubstyle_s
{
	int      id;
	fz_font *font;
	float    size;
	int      wmode;
	int      script;
	/* etc */
};

struct fz_epubstyles_s
{
        int            id;
	int            len, cap;
	fz_epubstyle **style;
};

typedef struct fz_epubrbox_s
{
	fz_point      origin;
	fz_point      dir;
	float         w;
	float         h;
	float         d;
}
fz_epubrbox;

static const fz_epubrbox fz_epubrbox_empty = { { 0.0,0.0 }, { 1.0,0.0 }, 0.0, 0.0, 0.0 };

typedef struct fz_epubchar_s
{
	fz_epubrbox box;
	int         c;
}
fz_epubchar;

typedef struct fz_epubword_s
{
	fz_epubstyle *style;
	int           len, cap;
	fz_epubchar  *chars;
}
fz_epubword;

typedef struct fz_epubline_s
{
	float         distance; /* Vertical distance from previous line */
	int           len, cap;
	fz_epubword **words;
}
fz_epubline;

typedef struct fz_epubstory_s
{
	int           len, cap;
	fz_epubline **lines;
}
fz_epubstory;

typedef struct fz_epubstories_s
{
	int            len, cap;
	fz_epubstory **story;
}
fz_epubstories;

struct fz_epubdevice_s
{
	fz_epub        *epub;
	fz_device      *dev;
	fz_epubstyles  *styles;
	fz_epubstories *stories;
};

/** Word functions *****************************************/

fz_epubword *
fz_epubword_create(fz_epubchar ch, fz_epubstyle *style)
{
	fz_epubword *word;
	word = fz_malloc(sizeof(*word));
	word->style    = style;
	word->len      = 1;
	word->cap      = 4;
	word->chars    = fz_malloc(word->cap * sizeof(*word->chars));
	word->chars[0] = ch;
	return word;
}

void
fz_epubword_destroy(fz_epubword *word)
{
	if (word == NULL)
		return;
	fz_free(word->chars);
	fz_free(word);
}

void
fz_epubword_appendchar(fz_epubword *word, fz_epubchar ch)
{
	assert(word != NULL);
	if (word->len == word->cap)
	{
		word->cap *= 2;
		word->chars = fz_realloc(word->chars, word->cap, sizeof(word->chars));
	}
	word->chars[word->len++] = ch;
}

void
fz_epubword_prependchar(fz_epubword *word, fz_epubchar ch)
{
	assert(word != NULL);
	if (word->len == word->cap)
	{
		word->cap *= 2;
		word->chars = fz_realloc(word->chars, word->cap, sizeof(*word->chars));
	}
	memmove(&word->chars[1], &word->chars[0], word->len *sizeof(*word->chars));
	word->chars[0] = ch;
	word->len++;
}

/** Line functions *****************************************/

fz_epubline *
fz_epubline_create(fz_epubword *word, float distance)
{
	fz_epubline *line;
	line = fz_malloc(sizeof(*line));
	line->distance = distance;
	line->len      = 1;
	line->cap      = 4;
	line->words    = fz_malloc(line->cap * sizeof(*line->words));
	line->words[0] = word;
	return line;
}

static void
fz_epubline_destroy(fz_epubline *line)
{
	int i;

	if (line == NULL)
		return;
	for (i = 0; i < line->len; i++)
	{
		fz_epubword_destroy(line->words[i]);
	}
	fz_free(line->words);
	fz_free(line);
}

static void
fz_epubline_appendword(fz_epubline *line, fz_epubword *word)
{
	assert(line != NULL);
	assert(word != NULL);
	if (line->len == line->cap)
	{
		line->cap *= 2;
		line->words = fz_realloc(line->words, line->cap, sizeof(*line->words));
	}
	line->words[line->len++] = word;
}

static void
fz_epubline_prependword(fz_epubline *line, fz_epubword *word)
{
	assert(line != NULL);
	assert(word != NULL);
	if (line->len == line->cap)
	{
		line->cap *= 2;
		line->words = fz_realloc(line->words, line->cap, sizeof(*line->words));
	}
	memmove(&line->words[1], &line->words[0], sizeof(*line->words)*line->len);
	line->words[0] = word;
	line->len++;
}

static void
fz_showline(fz_epubline *line)
{
	int word_n;

	if (line == NULL)
		return;

	for (word_n=0; word_n < line->len; word_n++)
	{
		fz_epubword *word = line->words[word_n];
		int char_n;
		for (char_n=0; char_n < word->len; char_n++)
		{
			int c = word->chars[char_n].c;
			if (c == '<')
				printf("&lt;");
			else if (c == '>')
				printf("&gt;");
			else if (c == '&')
				printf("&amp;");
			else if ((c >= 32) && (c <= 127))
				printf("%c", c);
			else
				printf("&#%d;", c);
		}
	}
}

/** Story functions *****************************************/

static fz_epubstory *
fz_epubstory_create(fz_epubline *line)
{
	fz_epubstory *story;
	story = fz_malloc(sizeof(*story));
	story->len      = 1;
	story->cap      = 4;
	story->lines    = fz_malloc(sizeof(*story->lines) * story->cap);
	story->lines[0] = line;
	return story;
}

static void
fz_epubstory_destroy(fz_epubstory *story)
{
	int i;

	if (story == NULL)
		return;
	for (i = 0; i < story->len; i++)
	{
		fz_epubline_destroy(story->lines[i]);
	}
	fz_free(story->lines);
	fz_free(story);
}

static void
fz_epubstory_appendline(fz_epubstory *story, fz_epubline *line)
{
	assert(story != NULL);
	assert(line != NULL);
	if (story->len == story->cap)
	{
		story->cap *= 2;
		story->lines = fz_realloc(story->lines, story->cap, sizeof(*story->lines));
	}
	story->lines[story->len++] = line;
}

static void
fz_epubstory_prependline(fz_epubstory *story, fz_epubline *line)
{
	assert(story != NULL);
	assert(line != NULL);
	if (story->len == story->cap)
	{
		story->cap *= 2;
		story->lines = fz_realloc(story->lines, story->cap, sizeof(*story->lines));
	}
	memmove(&story->lines[1], &story->lines[0], story->len*sizeof(*story->lines));
	story->lines[0] = line;
	story->len++;
}

/** Stories functions *****************************************/

static fz_epubstories *
fz_epubstories_create(void)
{
	fz_epubstories *stories;

	stories = fz_malloc(sizeof(*stories));
	stories->len   = 0;
	stories->cap   = 0;
	stories->story = NULL;
	return stories;
}

static void
fz_epubstories_destroy(fz_epubstories *stories)
{
	int i;

	if (stories == NULL)
		return;
	for (i = 0; i < stories->len; i++)
		fz_epubstory_destroy(stories->story[i]);
	fz_free(stories);
}

static void
fz_epubstories_appendstory(fz_epubstories *stories, fz_epubstory *story)
{
	assert(stories != NULL);
	assert(story != NULL);
	if (stories->len == stories->cap)
	{
		stories->cap *= 2;
		if (stories->cap == 0)
			stories->cap = 4;
		stories->story = fz_realloc(stories->story, stories->cap, sizeof(*stories->story));
	}
	stories->story[stories->len++] = story;
}

/** Style handling functions */

static fz_epubstyle *
do_find_style(fz_epubstyles *styles, float size, fz_font *font, int wmode, int script)
{
	int i;
	fz_epubstyle *style;

	assert(styles);
	for (i = 0; i < styles->len; i++)
	{
		style = styles->style[i];
		if ((style->font   == font) &&
		    (style->size   == size) &&
		    (style->wmode  == wmode) &&
		    (style->script == script)) /* FIXME: others */
		{
			break;
		}
	}
	if (i != styles->len)
		return style;

	/* Better make a new one */
	style = fz_malloc(sizeof(*style));
	style->id     = styles->id++;
	style->font   = fz_keepfont(font);
	style->size   = size;
	style->wmode  = wmode;
	style->script = script;
	/* And add it to our list */
	if (styles->len == styles->cap)
	{
		styles->cap *= 2;
		if (styles->cap == 0)
			styles->cap = 4;
		styles->style = fz_realloc(styles->style, styles->cap, sizeof(*styles->style));
	}
	styles->style[styles->len++] = style;
	return style;
}

fz_epubstyle *
fz_epub_findstyle(fz_epubstyles *styles, fz_text *text, fz_matrix *ctm,
	fz_colorspace *colorspace, float *color, float alpha, fz_strokestate *stroke)
{
	float size = 0.0f;
	fz_font *font = (text ? text->font : NULL);
	int wmode = (text ? text->wmode : 0);

	if (ctm && text)
	{
		fz_matrix tm = text->trm;
		fz_matrix trm;
		tm.e = 0;
		tm.f = 0;
		trm = fz_concat(tm, *ctm);
		size = fz_matrixexpansion(trm);
	}
	return do_find_style(styles, size, font, wmode, 0);
}

static fz_epubstyle *
fz_epubstyle_applyscriptdelta(fz_epubstyles *styles, fz_epubstyle *style, int delta)
{
	return do_find_style(styles, style->size, style->font, style->wmode, delta);
}

static void
fz_epubstyle_destroy(fz_epubstyle *style)
{
	if (style == NULL)
		return;
	fz_dropfont(style->font);
	fz_free(style);
}

int
fz_epubstyle_writestyle(fz_epub *epub, fz_epubstyle *style)
{
	char text[80];
	if (style == NULL)
		return 0;
	sprintf(text, "#style%x {\n  font-family:\"%s\";\n  font-size: %gem\n}\n", style->id, style->font->name, style->size);
	return PRINT2ZIP(epub, text);
}

static void
fz_epubstyle_start(fz_epub *epub, fz_epubstyle *style)
{
	char text[80];
	int script = style->script;

	sprintf(text, "<span class=\"style%x\">", style->id);
	PRINT2ZIP(epub, text);
	while (script > 0)
	{
		PRINT2ZIP(epub, "<sup>");
		script--;
	}
	while (script < 0)
	{
		PRINT2ZIP(epub, "<sub>");
		script++;
	}
}

static void
fz_epubstyle_stop(fz_epub *epub, fz_epubstyle *style)
{
	int script = style->script;

	while (script > 0)
	{
		PRINT2ZIP(epub, "</sup>");
		script--;
	}
	while (script < 0)
	{
		PRINT2ZIP(epub, "</sub>");
		script++;
	}
	PRINT2ZIP(epub, "</span>");
}

/** Styles functions */

fz_epubstyles *
fz_epubstyles_create(void)
{
	fz_epubstyles *styles;

	styles = fz_malloc(sizeof(*styles));
	styles->id    = 0;
	styles->len   = 0;
	styles->cap   = 0;
	styles->style = NULL;
	return styles;
}

void
fz_epubstyles_destroy(fz_epubstyles *styles)
{
	int i;

	if (styles == NULL)
		return;
	for (i = 0; i < styles->len; i++)
		fz_epubstyle_destroy(styles->style[i]);
	fz_free(styles);
}

int
fz_epubstyles_writestylesheet(fz_epub *epub)
{
	int i, e;

	if (epub == NULL || epub->styles == NULL)
		return 0;
	for (i = 0; i < epub->styles->len; i++)
	{
		e = fz_epubstyle_writestyle(epub, epub->styles->style[i]);
		if (e)
			break;
	}
	return e;
}

static fz_point
fz_epubrbox_end(fz_epubrbox box)
{
	fz_point end;
	end.x = box.origin.x + box.w * box.dir.x;
	end.y = box.origin.y + box.w * box.dir.y;
	return end;
}

enum
{
    match_none      = 0,
    match_forwards  = 1,
    match_backwards = 2,
    match_withspace = 4,
};

static void
fz_addtextcharimp(fz_epubdevice *edev, fz_epubstyle **style, float size, int c, fz_epubrbox rbox)
{
	int story_n;
	int match = match_none;
	fz_epubstory *story;
	fz_epubstory *story_continue = NULL;
	fz_epubline *line;
	fz_epubword *word;
	fz_epubchar *chr;
	fz_point delta, end, chend;
	float dist, dot, linedistance;
	fz_epubchar ch;
	int script_delta = 0;

	assert(edev);
	assert(edev->stories);

	ch.box = rbox;
	ch.c = c;
	chend = fz_epubrbox_end(rbox);
	DBUG(("Char: %c (origin=%g %g, end=%g %g)\n",
		ch.c, rbox.origin.x, rbox.origin.y, chend.x, chend.y));

	/* First, search the stories... */
	for (story_n = edev->stories->len-1; match == match_none && story_n >= 0; story_n--)
	{
		/* ...for a story... */
		int line_n;
		story = edev->stories->story[story_n];

		for (line_n = story->len-1; match == match_none && line_n >= 0; line_n--)
		{
			/* ...that contains a line... */
			line = story->lines[line_n];
			/* ...that ends in a word... */
			word = line->words[line->len-1];
			/* ...that ends at the right place... */
			chr = &word->chars[word->len-1];

			end = fz_epubrbox_end(chr->box);
			delta.x = rbox.origin.x - end.x;
			delta.y = rbox.origin.y - end.y;
			dist = sqrtf(delta.x * delta.x + delta.y * delta.y);
			DBUG(("Last char: %c ended at %g %g (delta=%g %g, dist %g)\n",
				chr->c, end.x, end.y, delta.x, delta.y, dist));
			if (dist < 0.1*size)
			{
				/* We'll accept a small difference in any direction as being a continuation of the line */
				match = match_forwards;
				DBUG(("Simple next char match\n"));
				break;
			}
			else
			{
				/* Let's consider the direction */
				fz_point ndelta;
				ndelta.x = delta.x / dist;
				ndelta.y = delta.y / dist;
				dot = ndelta.x * chr->box.dir.x + ndelta.y * chr->box.dir.y;
				if ((dot > 0.95f) && (dist <= 2*size))
				{
					/* It's in line, and not stupidly far away. */
					if ((dist > size * SPACEDIST) && (chr->c != ' ') && (c != ' '))
					{
						/* Throw in a space, and we'll take it! */
						DBUG(("+space = match\n"));
						match = match_forwards | match_withspace;
						break;
					}
					/* We'll take it as is */
					DBUG(("next char match\n"));
					match = match_forwards;
					break;
				}
				else if ((dot < -0.95) && (dist < 0.5*size))
				{
					/* Squashed in closer; We'll take it as is */
					DBUG(("next char match\n"));
					match = match_forwards;
					break;
				}
				if (dist < 0.75*size)
				{
					/* Could be a superscript or subscript? */
					dot = ndelta.x * chr->box.dir.y - ndelta.y * chr->box.dir.x;
					if (dot > 0.95f)
					{
						DBUG(("subscript!\n"));
						match = match_forwards;
						script_delta = -1;
						break;
					}
					else if (dot < -0.95f)
					{
						DBUG(("superscript!\n"));
						match = match_forwards;
						script_delta = 1;
						break;
					}
				}
				/* Consider a 'next line' match. */
				if (!story_continue)
				{
					linedistance = delta.x * chr->box.dir.y - delta.y * chr->box.dir.x;
					if ((linedistance >= .8*size) && (linedistance <= 2*size))
					{
						/* FIXME: Should check line direction is appropriate too */
						/* This seems plausibly the next line. */
						DBUG(("next line match maybe\n"));
						story_continue = story;
						break;
					}
				}
			}
			/* ...or that starts with a word... */
			word = line->words[0];
			/* ...that starts at the right place... */
			chr = &word->chars[0];
			delta.x = chr->box.origin.x - chend.x;
			delta.y = chr->box.origin.y - chend.y;
			dist = sqrtf(delta.x * delta.x + delta.y * delta.y);
			DBUG(("First char: %c starts at %g %g (delta=%g %g, dist %g\n",
				chr->c, chr->box.origin.x, chr->box.origin.y, delta.x, delta.y, dist));
			if (dist < 0.1*size)
			{
				/* We'll accept a small difference in any direction as being a continuation of the line */
				DBUG(("Simple match backwards\n"));
				match = match_backwards;
				break;
			}
			else
			{
				/* Let's consider the direction */
				fz_point ndelta;
				ndelta.x = delta.x / dist;
				ndelta.y = delta.y / dist;
				dot = ndelta.x * chr->box.dir.x + ndelta.y * chr->box.dir.y;
				if ((dot > 0.95f) && (dist <= 2*size))
				{
					/* It's in line, and not stupidly far away. */
					if ((dist > size * SPACEDIST) && (chr->c != ' ') && (c != ' '))
					{
						/* Throw in a space, and we'll take it! */
						DBUG(("+space match backwards\n"));
						match = match_backwards | match_withspace;
						break;
					}
					/* We'll take it as is */
					DBUG(("previous char match\n"));
					match = match_backwards;
					break;
				}
				else if ((dot < -.95) && (dist < 0.5*size))
				{
					/* Squashed in close, but we'll take it as is */
					DBUG(("previous char match\n"));
					match = match_backwards;
					break;
				}
			}
		}
	}
	if (script_delta)
	{
		*style = fz_epubstyle_applyscriptdelta(edev->epub->styles, *style, word->style->script + script_delta);
	}
	if (match & match_backwards)
	{
		if (match == match_withspace)
		{
			/* Add a space to the current word */
			fz_epubchar sch;
			sch.c = ' ';
			sch.box          = rbox;
			sch.box.origin.x = end.x;
			sch.box.origin.y = end.x;
			sch.box.w        = dist;
			fz_epubword_prependchar(word, sch);
		}
		/* Does the style match? */
		if (word->style == *style)
		{
			/* Just add the char to the current word */
			fz_epubword_prependchar(word, ch);
		}
		else
		{
			/* Add a new word */
			word = fz_epubword_create(ch, *style);
			fz_epubline_prependword(line, word);
		}
	}
	else if (match & match_forwards)
	{
		if (match & match_withspace)
		{
			/* Add a space to the current word */
			fz_epubchar sch;
			sch.c = ' ';
			sch.box          = rbox;
			sch.box.origin.x = end.x;
			sch.box.origin.y = end.x;
			sch.box.w        = dist;
			fz_epubword_appendchar(word, sch);
		}
		/* Does the style match? */
		if (word->style == *style)
		{
			/* Just add the char to the current word */
			fz_epubword_appendchar(word, ch);
		}
		else
		{
			/* Add a new word */
			word = fz_epubword_create(ch, *style);
			fz_epubline_appendword(line, word);
		}
	}
	else if (story_continue)
	{
		/* Add a new line */
		word = fz_epubword_create(ch, *style);
		line = fz_epubline_create(word, linedistance);
		fz_epubstory_appendline(story_continue, line);
	}
	else
	{
		/* Add it as a singleton */
		word = fz_epubword_create(ch, *style);
		line = fz_epubline_create(word, 0.0f);
		story = fz_epubstory_create(line);
		fz_epubstories_appendstory(edev->stories, story);
	}
}

static fz_epubrbox
fz_splitrbox(fz_epubrbox rbox, int i, int n)
{
	fz_epubrbox rbox2 = rbox;
	rbox2.w /= n;
	rbox2.origin.x += rbox2.dir.x * rbox2.w * i;
	rbox2.origin.y += rbox2.dir.y * rbox2.w * i;
	return rbox2;
}

static void
fz_addtextchar(fz_epubdevice *edev, fz_epubstyle **style, float size, int c, fz_epubrbox rbox)
{
	switch (c)
	{
	case -1: /* ignore when one unicode character maps to multiple glyphs */
		break;
	case 0xFB00: /* ff */
		fz_addtextcharimp(edev, style, size, 'f', fz_splitrbox(rbox, 0, 2));
		fz_addtextcharimp(edev, style, size, 'f', fz_splitrbox(rbox, 1, 2));
		break;
	case 0xFB01: /* fi */
		fz_addtextcharimp(edev, style, size, 'f', fz_splitrbox(rbox, 0, 2));
		fz_addtextcharimp(edev, style, size, 'i', fz_splitrbox(rbox, 1, 2));
		break;
	case 0xFB02: /* fl */
		fz_addtextcharimp(edev, style, size, 'f', fz_splitrbox(rbox, 0, 2));
		fz_addtextcharimp(edev, style, size, 'l', fz_splitrbox(rbox, 1, 2));
		break;
	case 0xFB03: /* ffi */
		fz_addtextcharimp(edev, style, size, 'f', fz_splitrbox(rbox, 0, 3));
		fz_addtextcharimp(edev, style, size, 'f', fz_splitrbox(rbox, 1, 3));
		fz_addtextcharimp(edev, style, size, 'i', fz_splitrbox(rbox, 2, 3));
		break;
	case 0xFB04: /* ffl */
		fz_addtextcharimp(edev, style, size, 'f', fz_splitrbox(rbox, 0, 3));
		fz_addtextcharimp(edev, style, size, 'f', fz_splitrbox(rbox, 1, 3));
		fz_addtextcharimp(edev, style, size, 'l', fz_splitrbox(rbox, 2, 3));
		break;
	case 0xFB05: /* long st */
	case 0xFB06: /* st */
		fz_addtextcharimp(edev, style, size, 's', fz_splitrbox(rbox, 0, 2));
		fz_addtextcharimp(edev, style, size, 't', fz_splitrbox(rbox, 1, 2));
		break;
	default:
		fz_addtextcharimp(edev, style, size, c, rbox);
		break;
	}
}

static void
fz_epub_handlechars(fz_epubdevice *edev, fz_text *text, fz_matrix ctm, fz_epubstyle *style)
{
	fz_font *font = text->font;
	FT_Face face = font->ftface;
	fz_matrix tm = text->trm;
	fz_matrix trm;
	float size;
	float adv;
	fz_point dir, ndir;
	float dist;
	float ascender = 1;
	float descender = 0;
	int multi;
	int i, j, err;
	fz_epubrbox rbox;
	fz_point factor;

	if (text->len == 0)
		return;

	if (font->ftface)
	{
		err = FT_Set_Char_Size(font->ftface, 64, 64, 72, 72);
		if (err)
			fz_warn("freetype set character size: %s", ft_errorstring(err));
		ascender = (float)face->ascender / face->units_per_EM;
		descender = (float)face->descender / face->units_per_EM;
	}

	rbox = fz_epubrbox_empty;

	if (text->wmode == 0)
	{
		dir.x = 1;
		dir.y = 0;
	}
	else
	{
		dir.x = 0;
		dir.y = -1;
	}

	tm.e = 0;
	tm.f = 0;
	trm = fz_concat(tm, ctm);
	dir = fz_transformvector(trm, dir);
	dist = sqrtf(dir.x * dir.x + dir.y * dir.y);
	ndir.x = dir.x / dist;
	ndir.y = dir.y / dist;

	rbox.dir = ndir;

	size = fz_matrixexpansion(trm);
	factor = fz_matrixscalefactors(trm);

	multi = 1;

	for (i = 0; i < text->len; i++)
	{
		tm.e = text->els[i].x;
		tm.f = text->els[i].y;
		trm = fz_concat(tm, ctm);
		/* Calculate bounding box and new pen position based on font metrics */
		if (font->ftface)
		{
			FT_Fixed ftadv = 0;
			int mask = FT_LOAD_NO_BITMAP | FT_LOAD_NO_HINTING | FT_LOAD_IGNORE_TRANSFORM;
			if (text->wmode)
				mask |= FT_LOAD_VERTICAL_LAYOUT;
			FT_Get_Advance(font->ftface, text->els[i].gid, mask, &ftadv);
			adv = ftadv / 65536.0f;
			if (text->wmode)
			{
				//adv = -1; /* TODO: freetype returns broken vertical metrics */
				//rect.x0 = 0;
				//rect.y0 = 0;
				//rect.x1 = 1;
				//rect.y1 = adv;
				rbox.w = factor.y;  // Should be based on adv
				rbox.h = 0;
				rbox.d = factor.x;
			}
			else
			{
				//rect.x0 = 0;
				//rect.y0 = descender;
				//rect.x1 = adv;
				//rect.y1 = ascender;
				rbox.w = adv*factor.x;
				rbox.h = ascender*factor.y;
				rbox.d = descender*factor.y;
			}
		}
		else
		{
			adv = font->t3widths[text->els[i].gid];
			//rect.x0 = 0;
			//rect.y0 = descender;
			//rect.x1 = adv;
			//rect.y1 = ascender;
			rbox.w = adv*factor.x;
			rbox.h = ascender*factor.y;
			rbox.d = descender*factor.y;
		}
		rbox.origin.x = trm.e;
		rbox.origin.y = trm.f;
		rbox.dir = ndir;
		/* See if this is a multi-part char */
		for (j = i+1; j < text->len; j++)
		{
			if (text->els[j].gid >= 0)
				break;
		}
		multi = j-i;
		if (multi == 1)
		{
			fz_addtextchar(edev, &style, size, text->els[i].ucs, rbox);
		}
		else
		{
			fz_epubrbox rbox2 = rbox;
			rbox2.w = rbox.w/multi;
			for (j = 0; j < multi; j++)
			{
				fz_addtextchar(edev, &style, size, text->els[i].ucs, rbox2);
				rbox2.origin.x += rbox2.w * rbox.dir.x;
				rbox2.origin.y += rbox2.w * rbox.dir.y;
			}
		}
	}
}

static void
fz_epubfilltext(void *user, fz_text *text, fz_matrix ctm,
	fz_colorspace *colorspace, float *color, float alpha)
{
	fz_epubdevice *edev = user;
	fz_epubstyle *style;

	style = fz_epub_findstyle(edev->epub->styles, text, &ctm, colorspace, color, alpha, NULL);
	fz_epub_handlechars(edev, text, ctm, style);
}

static void
fz_epubstroketext(void *user, fz_text *text, fz_strokestate *stroke, fz_matrix ctm,
	fz_colorspace *colorspace, float *color, float alpha)
{
	fz_epubdevice *edev = user;
	fz_epubstyle *style;

	style = fz_epub_findstyle(edev->epub->styles, text, &ctm, colorspace, color, alpha, stroke);
	fz_epub_handlechars(edev, text, ctm, style);
}

static void
fz_epubcliptext(void *user, fz_text *text, fz_matrix ctm, int accumulate)
{
	fz_epubdevice *edev = user;
	fz_epubstyle *style;

	style = fz_epub_findstyle(edev->epub->styles, text, &ctm, NULL, NULL, 0, NULL);
	fz_epub_handlechars(edev, text, ctm, style);
}

static void
fz_epubclipstroketext(void *user, fz_text *text, fz_strokestate *stroke, fz_matrix ctm)
{
	fz_epubdevice *edev = user;
	fz_epubstyle *style;

	style = fz_epub_findstyle(edev->epub->styles, text, &ctm, NULL, NULL, 0, stroke);
	fz_epub_handlechars(edev, text, ctm, style);
}

static void
fz_epubignoretext(void *user, fz_text *text, fz_matrix ctm)
{
	fz_epubdevice *edev = user;
	fz_epubstyle *style;

	style = fz_epub_findstyle(edev->epub->styles, text, NULL, NULL, NULL, 0, NULL);
	fz_epub_handlechars(edev, text, ctm, style);
}

static void
analyseStories(fz_epubstories *stories)
{
	int story_n;

	if (stories == NULL)
		return;

	for (story_n = 0; story_n < stories->len; story_n++)
	{
		fz_epubstory *story = stories->story[story_n];
		fz_epubline *line;
		fz_epubword *word;
		int line_n;
		fz_point l, r;
		fz_point perp;
		float w;

		/* Find the bounds for the lines in this story */
		line  = story->lines[0];
		l     = line->words[0]->chars[0].box.origin;
		perp  = line->words[0]->chars[0].box.dir;
		word  = line->words[line->len-1];
		r     = word->chars[word->len-1].box.origin;
		w     = word->chars[word->len-1].box.w;
		r.x  += word->chars[word->len-1].box.dir.x * w;
		r.y  += word->chars[word->len-1].box.dir.y * w;
		for (line_n = 1; line_n < story->len; line_n++)
		{
			fz_point p, delta;
			float distance;

			line = story->lines[line_n];
			/* Compare left hand edge */
			p        = line->words[0]->chars[0].box.origin;
			delta.x  = l.x - p.x;
			delta.y  = l.y - p.y;
			distance = delta.x * perp.x + delta.y * perp.y;
			if (distance < -line->words[0]->style->size)
			{
				/* This line probably starts a new paragraph */
				/* This goes wrong with lists or hanging indents */
				line->distance = 0;
			}
			else
			{
				l = p;
			}
			/* Compare right hand edge */
			word     = line->words[line->len-1];
			p        = word->chars[word->len-1].box.origin;
			w        = word->chars[word->len-1].box.w;
			p.x     += word->chars[word->len-1].box.dir.x * w;
			p.y     += word->chars[word->len-1].box.dir.y * w;
			delta.x  = r.x - p.x;
			delta.y  = r.y - p.y;
			distance = delta.x * perp.x + delta.y * perp.y;
			if (distance > word->style->size)
			{
			    /* This line probably ends a paragraph */
			    /* This will go wrong with ragged right text */
			    if (line_n+1 < story->len)
			    {
				    story->lines[line_n+1]->distance = 0;
			    }
			}
			else
			{
			    r = p;
			}
		}
	}
}

static int
fz_hyphenatedlines(fz_epubline *line1, fz_epubline *line2)
{
	fz_epubword *word;
	int c = line2->words[0]->chars[0].c;
	/* If second line doesn't start with an alpha char, it can't be hyphenated */
	if (!(((c >= 'a') && (c <= 'z')) || ((c >= 'A') && (c <= 'Z'))))
		return 0;
	/* If the first line doesn't end in a '-' it can't be hyphenated */
	word = line1->words[line1->len-1];
	if (word->chars[word->len-1].c != '-')
		return 0;
	/* The character before must be alpha */
	if (word->len > 1)
	{
		c = word->chars[word->len-2].c;
		return (((c >= 'a') && (c <= 'z')) || ((c >= 'A') && (c <= 'Z')));
	}
	if (line1->len <= 1)
	{
		/* There is no character before! */
		return 0;
	}
	word = line1->words[line1->len-2];
	c = word->chars[word->len-1].c;
	return (((c >= 'a') && (c <= 'z')) || ((c >= 'A') && (c <= 'Z')));
}

static int
fz_spaceterminatedline(fz_epubline *line)
{
	fz_epubword *word = line->words[line->len-1];
	return word->chars[word->len-1].c == ' ';
}

static void
amalgamateLines(fz_epubstories *stories)
{
	int story_n;

	if (stories == NULL)
		return;

	for (story_n = 0; story_n < stories->len; story_n++)
	{
		fz_epubstory *story = stories->story[story_n];
		fz_epubline *line_w, *line_r;
		fz_epubword *word, *word2;
		int line_w_n, line_r_n, word_n, i;

		line_w_n = 0;
		for (line_r_n = 1; line_r_n < story->len; line_r_n++)
		{
			line_w = story->lines[line_w_n];
			line_r = story->lines[line_r_n];
			if (line_r->distance == 0)
			{
			    /* New paragraph */
			    line_w_n++;
			    if (line_w_n != line_r_n)
			    {
				    story->lines[line_w_n] = line_r;
       				story->lines[line_r_n] = NULL;
			    }
			    continue;
			}
			/* So we're going to amalgamate the lines */
			if (fz_hyphenatedlines(line_w, line_r))
			{
				/* Remove the last - */
				word = line_w->words[line_w->len-1];
				word->len--;
				if (word->len == 0)
				{
					/* Remove the word. This is safe. */
					fz_epubword_destroy(word);
					line_w->len--;
				}
			}
			else if (!fz_spaceterminatedline(line_w))
			{
				/* Add a space */
				fz_epubchar ch;
				word = line_w->words[line_w->len-1];
				/* ch.rbox isn't right here, but we don't care at this stage */
				ch = word->chars[word->len-1];
				ch.c = ' ';
				fz_epubword_appendchar(word, ch);
			}
			/* Copy line_r to line_w */
			word_n = 0;
			word   = line_w->words[line_w->len-1];
			word2  = line_r->words[0];
			if (word->style == word2->style)
			{
				/* If word styles match, copy chars. */
				for (i=0; i < word2->len; i++)
				{
					fz_epubword_appendchar(word, word2->chars[i]);
				}
				fz_epubword_destroy(word2);
				line_r->words[0] = NULL;
				word_n++;
			}
			for (;word_n < line_r->len; word_n++)
			{
				fz_epubline_appendword(line_w, line_r->words[word_n]);
				line_r->words[word_n] = NULL;
			}
			fz_epubline_destroy(line_r);
			story->lines[line_r_n] = NULL;
		}
		story->len = line_w_n+1;
	}
}


static void
fz_epubfreeuser(void *user)
{
	char text[40];
	fz_epubdevice *edev = user;

	analyseStories(edev->stories);

	amalgamateLines(edev->stories);

	/* Now we dump the whole page */
	if (edev->stories != NULL)
	{
		int story_n;
		fz_epubstyle *style = NULL;
		for (story_n = 0; story_n < edev->stories->len; story_n++)
		{
			fz_epubstory *story = edev->stories->story[story_n];
			int line_n;
			for (line_n = 0; line_n < story->len; line_n++)
			{
				fz_epubline *line = story->lines[line_n];
				int word_n;
				sprintf(text, "<p>");
				PRINT2ZIP(edev->epub, text);
				style = NULL;

				for (word_n = 0; word_n < line->len; word_n++)
				{
					fz_epubword *word = line->words[word_n];
					int ch_n;
					if (style != word->style)
					{
						if (style != NULL)
							fz_epubstyle_stop(edev->epub, style);
						fz_epubstyle_start(edev->epub, word->style);
						style = word->style;
					}
					for (ch_n = 0; ch_n < word->len; ch_n++)
					{
						fz_epubchar *ch = &word->chars[ch_n];
						if (ch->c == '<')
							sprintf(text, "&lt;");
						else if (ch->c == '>')
							sprintf(text, "&gt;");
						else if (ch->c == '&')
							sprintf(text, "&amp;");
						else if ((ch->c >= 32) && (ch->c <= 127))
							sprintf(text, "%c", ch->c);
						else
							sprintf(text, "&#%d;", ch->c);
						PRINT2ZIP(edev->epub, text);
					}
				}
				fz_epubstyle_stop(edev->epub, style);
				sprintf(text, "</p>\n");
				PRINT2ZIP(edev->epub, text);
			}
		}
	}

	/* TODO: unicode NFC normalization */
	/* TODO: bidi logical reordering */

	/* FIXME: This is where we should do the output */

	fz_free(edev);
}

fz_device *
fz_newepubdevice(fz_epub *epub)
{
	fz_device *dev;
	fz_epubdevice *edev = fz_malloc(sizeof(*edev));
	edev->epub = epub;
	edev->stories = fz_epubstories_create();

	dev = fz_newdevice(edev);
	dev->hints          = FZ_IGNOREIMAGE | FZ_IGNORESHADE;
	dev->freeuser       = fz_epubfreeuser;
	dev->filltext       = fz_epubfilltext;
	dev->stroketext     = fz_epubstroketext;
	dev->cliptext       = fz_epubcliptext;
	dev->clipstroketext = fz_epubclipstroketext;
	dev->ignoretext     = fz_epubignoretext;
	return dev;
}
