#include "fitz.h"
#include "dev_epub.h"

static fz_error doWrite(fz_epub *epub, const char *filename, const char *data, unsigned int datalen)
{
	int e;

	e = zipOpenNewFileInZip(epub->outzip,
			filename,
			NULL, NULL, 0, NULL, 0, NULL,
			Z_DEFLATED, Z_BEST_COMPRESSION);
	if (e == 0)
		e = zipWriteInFileInZip(epub->outzip, data, datalen);
	if (e == 0)
		e = zipCloseFileInZip(epub->outzip);

	return (e == 0 ? fz_okay : fz_throw("Write into zipfile failed"));
}

static fz_error openChapter(fz_epub *epub)
{
	int e;
	char text[40];

	epub->chapterCount++;
	sprintf(text, "OEBPS/chap%02d.xhtml", epub->chapterCount);
	e = zipOpenNewFileInZip(epub->outzip,
			text,
			NULL, NULL, 0, NULL, 0, NULL,
			Z_DEFLATED, Z_BEST_COMPRESSION);

#define CHAPTERHEADING \
	"﻿<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>\n"\
	"<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.1//EN\" "\
	"\"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd\">\n"\
	"<html xmlns=\"http://www.w3.org/1999/xhtml\">\n"\
	"<head>\n"\
	"<link href=\"stylesheet.css\" type=\"text/css\" rel=\"stylesheet\" />\n"\
	"<link rel=\"stylesheet\" type=\"application/vnd.adobe-page-template+xml\" href=\"page-template.xpgt\"/>\n"\
	"<title>\n"
	DO(e,PRINT2ZIP(epub, CHAPTERHEADING));
	sprintf(text, "Chapter %d", epub->chapterCount);
	DO(e,PRINT2ZIP(epub, text));
	DO(e,PRINT2ZIP(epub, "</title></head><body>\n"));

	if (e)
		return fz_throw("Failed to open chapter");
	epub->chapterOpen = 1;

	return fz_okay;
}

static fz_error closeChapter(fz_epub *epub)
{
	int e;

	if (!epub->chapterOpen)
		return fz_okay;

	e = PRINT2ZIP(epub, "</body></html>\n");
	/* Close existing chapter file */
	DO(e, zipCloseFileInZip(epub->outzip));

	epub->chapterOpen = 0;
	if (e)
		return fz_throw("Error closing chapter");
	return fz_okay;
}

fz_error fz_epub_init(const char *filename, fz_epub **epubp)
{
	fz_epub  *epub;
	fz_error  error;

	epub = fz_malloc(sizeof(*epub));
	memset(epub, 0, sizeof(*epub));
	epub->outzip = zipOpen2(filename, 0, NULL, NULL);

	if (epub->outzip == NULL)
	{
		error = fz_throw("Failed to open zipFile %s", filename);
		goto cleanup;
	}
	epub->styles = fz_epubstyles_create();

	/* Output Mimetype file. Always the same. */
#define MIMETYPE "application/epub+zip"
	error = doWrite(epub, "mimetype", MIMETYPE, strlen(MIMETYPE));
	if (error)
		goto cleanup;

	/* Output Container file. Always the same. */
#define CONTAINER \
"<?xml version=\"1.0\"?>\n"\
"<container version=\"1.0\" xmlns=\"urn:oasis:names:tc:opendocument:xmlns:container\">\n"\
"  <rootfiles>\n"\
"    <rootfile full-path=\"OEBPS/content.opf\" media-type=\"application/oebps-package+xml\" />\n"\
"  </rootfiles>\n"\
"</container>\n"

	error = doWrite(epub, "META-INF/container.xml", CONTAINER, strlen(CONTAINER));
	if (error)
		goto cleanup;

	*epubp = epub;
	return fz_okay;

cleanup:
	*epubp = NULL;

	return error;
}

fz_error fz_epub_fin(fz_epub *epub)
{
	fz_error error;
	int e, i;
	char text[120];

	if (epub == NULL)
		return fz_okay;

	error = closeChapter(epub);

	/* Output content.opf file */
	e = zipOpenNewFileInZip(epub->outzip,
			"OEBPS/content.opf",
			NULL, NULL, 0, NULL, 0, NULL,
			Z_DEFLATED, Z_BEST_COMPRESSION);

#define CONTENTHEADER \
"﻿<?xml version=\"1.0\"?>\n"\
"<package version=\"2.0\" xmlns=\"http://www.idpf.org/2007/opf\"\n"\
"\tunique-identifier=\"BookId\">\n"\
" <metadata xmlns:dc=\"http://purl.org/dc/elements/1.1/\"\n"\
"           xmlns:opf=\"http://www.idpf.org/2007/opf\">\n"
	DO(e,PRINT2ZIP(epub, CONTENTHEADER));

	DO(e,PRINT2ZIP(epub, "  <dc:title>"));
	DO(e,PRINT2ZIP(epub, "MuPDF produced Ebook"));
	DO(e,PRINT2ZIP(epub, "</dc:title>\n"));

	DO(e,PRINT2ZIP(epub, "  <dc:creator opf:role=\"aut\">"));
	DO(e,PRINT2ZIP(epub, "MuPDF"));
	DO(e,PRINT2ZIP(epub, "</dc:creator>\n"));

	DO(e,PRINT2ZIP(epub, "  <dc:language>"));
	DO(e,PRINT2ZIP(epub, "en-GB"));
	DO(e,PRINT2ZIP(epub, "</dc:language>\n"));

	DO(e,PRINT2ZIP(epub, "  <dc:rights>"));
	DO(e,PRINT2ZIP(epub, "Public Domain"));
	DO(e,PRINT2ZIP(epub, "</dc:rights>\n"));

	DO(e,PRINT2ZIP(epub, "  <dc:publisher>"));
	DO(e,PRINT2ZIP(epub, ""));
	DO(e,PRINT2ZIP(epub, "</dc:publisher>\n"));

	DO(e,PRINT2ZIP(epub, "  <dc:identifier id=\"BookId\">"));
	DO(e,PRINT2ZIP(epub, ""));
	DO(e,PRINT2ZIP(epub, "</dc:identifier>\n"));

	DO(e,PRINT2ZIP(epub, " </metadata>\n"));
	DO(e,PRINT2ZIP(epub, " <manifest>\n"));

	DO(e,PRINT2ZIP(epub, "  <item id=\"ncx\" href=\"toc.ncx\" media-type=\"application/x-dtbncx+xml\" />\n"));
	DO(e,PRINT2ZIP(epub, "  <item id=\"style\" href=\"stylesheet.css\" media-type=\"text/css\" />\n"));
	for (i=1; i <= epub->chapterCount; i++)
	{
		sprintf(text, "  <item id=\"chapter%02d\" href=\"chap%02d.xhtml\" media-type=\"application/xhtml+xml\" />\n", i, i);
	        DO(e,PRINT2ZIP(epub, text));
	}
	DO(e,PRINT2ZIP(epub, " </manifest>\n"));
	DO(e,PRINT2ZIP(epub, " <spine toc=\"ncx\">\n"));
	DO(e,PRINT2ZIP(epub, "  <itemref idref=\"chapter01\" />\n"));
	DO(e,PRINT2ZIP(epub, " </spine>\n"));
	DO(e,PRINT2ZIP(epub, "</package>\n"));

	DO(e,zipCloseFileInZip(epub->outzip));

	/* Write toc.ncx file */
	DO(e,zipOpenNewFileInZip(epub->outzip,
			"OEBPS/toc.ncx",
			NULL, NULL, 0, NULL, 0, NULL,
			Z_DEFLATED, Z_BEST_COMPRESSION));
#define TOCHEADER \
"<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"\
"<ncx xmlns=\"http://www.daisy.org/z3986/2005/ncx/\" version=\"2005-1\">\n"\
"  <head>\n"\
"    <meta name=\"dtb:uid\" content=\"mupdf.com20101103133847\"/>\n"\
"    <meta name=\"dtb:depth\" content=\"1\"/>\n"\
"    <meta name=\"dtb:totalPageCount\" content=\"0\"/>\n"\
"    <meta name=\"dtb:maxPageNumber\" content=\"0\"/>\n"\
"  </head>\n"\
"  <docTitle>\n"\
"    <text>Mupdf generated .epub eBook</text>\n"\
"  </docTitle>\n"\
"  <navMap>\n"

#define TOCTRAILER \
"  </navMap>\n"\
"</ncx>"

	DO(e,PRINT2ZIP(epub, TOCHEADER));
	for (i=1; i <= epub->chapterCount; i++)
	{
		sprintf(text, "    <navPoint id=\"chapter%02d\" playOrder=\"%d\">\n", i, i);
		DO(e,PRINT2ZIP(epub, text));
		sprintf(text,"      <navLabel>\n        <text>Chapter %d</text>\n      </navLabel>\n", i);
		DO(e,PRINT2ZIP(epub, text));
		sprintf(text,"      <content src=\"chap01.xhtml\"/>\n    </navPoint>\n");
		DO(e,PRINT2ZIP(epub, text));
	}
	DO(e,PRINT2ZIP(epub, TOCTRAILER));
	DO(e,zipCloseFileInZip(epub->outzip));

	/* Write the stylesheet */
	DO(e,zipOpenNewFileInZip(epub->outzip,
			"OEBPS/stylesheet.css",
			NULL, NULL, 0, NULL, 0, NULL,
			Z_DEFLATED, Z_BEST_COMPRESSION));
	DO(e,fz_epubstyles_writestylesheet(epub));
	DO(e,zipCloseFileInZip(epub->outzip));

	DO(e,zipClose(epub->outzip, NULL));

	fz_epubstyles_destroy(epub->styles);
	fz_free(epub);
	if ((!error) && (e))
		error = fz_throw("Error whilst closing zipfile\n");
	return error;
}

fz_error fz_epub_outputpage(fz_epub *epub, fz_displaylist *list)
{
	fz_device *dev;
	epub->pageCount++;
	/* Ensure we have a chapter open */
	if (!epub->chapterOpen)
	{
		int error;
		error = openChapter(epub);
		if (error)
			return error;
	}

	dev = fz_newepubdevice(epub);
	fz_executedisplaylist(list, dev, fz_identity);
	fz_freedevice(dev);

	return fz_okay;
}
